/*
 * OSDKeys - zobrazenie stavu klvesov Lock a hlasitosti na obrazovke.
 * Copyright (C) 2008 Mat Sulr
 *
 * You can run this software, redistribute it and/or modify its source code.
 * Commercial use is prohibited (because of the limitation given by the
 * program used to create this application).
 *
 * This software is distributed WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you are unsure about legal details, see C++ Builder 6 Personal License
 * Terms by Borland Software Corporation.
 */

//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#include "osd.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma link "trayicon"
#pragma resource "*.dfm"

TfrmOsd *frmOsd; // formular zobrazujuci sa pri stlaceni danych klavesov
DWORD sprava_inst; // sprava, ktoru dostaneme po spusteni druhej instancie
void *stara_proc; // kvoli instancii
HWND predch_okno; // handle predchadzajuceho aktivneho okna

struct klaves {
    unsigned char kod; // Virtual-key code
    char retazec[20]; // parameter pre tvorbu "atomu"
    ATOM atom; // vratena unikatna hodnota, pouziva sa pri klaves. skratkach
};

// pole struktur s informaciami o klavesoch (kod, retazec, atom)
klaves klavesy[] = {
 {VK_CAPITAL, "OSDKeys CapsLock", 0},
 {VK_NUMLOCK, "OSDKeys NumLock", 0},
 {VK_SCROLL, "OSDKeys ScrollLock", 0},
 {VK_VOLUME_MUTE, "OSDKeys hlas stlm", 0},
 {VK_VOLUME_DOWN, "OSDKeys hlas minus", 0},
 {VK_VOLUME_UP, "OSDKeys hlas plus", 0}
};

const unsigned char POCET_KLAVESOV = sizeof(klavesy) / sizeof(klaves);

// parameter funkcie zisti_hlas
enum {
    HLAS_HLASITOST,
    HLAS_STLMENIE
};

// oznamovacie texty
const char *CAPS_ZAP = "CapsLock On";
const char *CAPS_VYP = "CapsLock Off";
const char *NUM_ZAP = "NumLock On";
const char *NUM_VYP = "NumLock Off";
const char *SCROLL_ZAP = "ScrollLock On";
const char *SCROLL_VYP = "ScrollLock Off";
const char *STLM_ZAP = "Mute On";
const char *STLM_VYP = "Mute Off";
const char *HLASITOST = "Volume ";
//---------------------------------------------------------------------------

/* simuluje stlacenie tlacidla tak, ze sa prevedie bezna akcia (napr. stisenie
 hlasitosti), pricom nedostaneme spravu o klavesovej skratke */
void Stlac(unsigned char kod, unsigned int atom) {
    UnregisterHotKey(NULL, atom); // docasne zrusenie skratky
    keybd_event(kod, 0, 0, 0); // simulacia stlacenia klavesy
    keybd_event(kod, 0, KEYEVENTF_KEYUP, 0); // simulacia pustenia klavesy
    RegisterHotKey(NULL, atom, 0, kod); // skratku opat zaregistrujeme
}
//---------------------------------------------------------------------------

/* zisti, ci je zapnuty *Lock a vrati odpovedajuci informacny retazec */
String InfoLock(unsigned char kod) {
    switch (kod) {
        case VK_CAPITAL:
            return GetKeyState(VK_CAPITAL) ? CAPS_ZAP : CAPS_VYP;
        case VK_NUMLOCK:
            return GetKeyState(VK_NUMLOCK) ? NUM_ZAP : NUM_VYP;
        case VK_SCROLL:
            return GetKeyState(VK_SCROLL) ? SCROLL_ZAP : SCROLL_VYP;
        default:
            return "";
    }
}
//---------------------------------------------------------------------------

/* vrati pocet percent z cisla s rozsahom od 0 do 65535 */
unsigned char Percenta(unsigned short cislo) {
    return (float) cislo / 65535 * 100;
}

/* ak typ = HLAS_HLASITOST: vrati hlasitost "Volume Control" v percentach;
 ak typ = HLAS_STLMENIE: vrati 1, ak je hlas stlmeny, inak 0 */
unsigned char ZistiHlas(unsigned char typ) {
    // otvorenie zvukoveho mixera
    HMIXER mixer;
    mixerOpen(&mixer, MIXER_OBJECTF_MIXER, 0, 0, 0);

    // mixer line - reproduktory
    MIXERLINE line = {0};
    line.cbStruct = sizeof(MIXERLINE);
    line.dwComponentType = MIXERLINE_COMPONENTTYPE_DST_SPEAKERS;
    mixerGetLineInfo((HMIXEROBJ) mixer, &line,
     MIXER_GETLINEINFOF_COMPONENTTYPE);

    // ovladanie hlasitosti
    MIXERLINECONTROLS mlc = {0};
    MIXERCONTROL mc = {0};
    mlc.cbStruct = sizeof(MIXERLINECONTROLS);
    mlc.dwLineID = line.dwLineID;
    mlc.dwControlType = (typ == HLAS_HLASITOST) ?
     MIXERCONTROL_CONTROLTYPE_VOLUME : MIXERCONTROL_CONTROLTYPE_MUTE;
    mlc.cControls = 1;
    mlc.pamxctrl = &mc;
    mlc.cbmxctrl = sizeof(MIXERCONTROL);
    mixerGetLineControls((HMIXEROBJ) mixer, &mlc,
     MIXER_GETLINECONTROLSF_ONEBYTYPE);

    // uroven hlasitosti
    MIXERCONTROLDETAILS mcd = {0};
    MIXERCONTROLDETAILS_UNSIGNED mcdu = {0};
    MIXERCONTROLDETAILS_BOOLEAN mcb = {0};
    mcd.cbStruct = sizeof(MIXERCONTROLDETAILS);
    mcd.hwndOwner = 0;
    mcd.dwControlID = mc.dwControlID;
    if (typ == HLAS_HLASITOST)
        mcd.paDetails = &mcdu;
    else
        mcd.paDetails = &mcb;
    mcd.cbDetails = (typ == HLAS_HLASITOST) ?
     sizeof(MIXERCONTROLDETAILS_UNSIGNED) : sizeof(MIXERCONTROLDETAILS_BOOLEAN);
    mcd.cChannels = 1;
    mixerGetControlDetails((HMIXEROBJ) mixer, &mcd,
     MIXER_GETCONTROLDETAILSF_VALUE);

    return ((typ == HLAS_HLASITOST) ? Percenta(mcdu.dwValue) : mcb.fValue);
}
//---------------------------------------------------------------------------

__fastcall TfrmOsd::TfrmOsd(TComponent* Owner)
    : TForm(Owner)
{
    predch_okno = GetForegroundWindow();
    Application->ShowMainForm = false;
}
//---------------------------------------------------------------------------

void __fastcall TfrmOsd::FormCreate(TObject *Sender)
{
    // skrytie aplikacie z panela uloh
    SetWindowLong(Application->Handle, GWL_EXSTYLE, WS_EX_TOOLWINDOW);
    
    frmOsd->DoubleBuffered = true;
    frmOsd->Left = 5;
    frmOsd->Top = 5;

    // registracia klavesovych skratiek
    for (unsigned char i = 0; i < POCET_KLAVESOV; i++) {
        klavesy[i].atom = GlobalAddAtom(klavesy[i].retazec);
        RegisterHotKey(NULL, klavesy[i].atom, 0, klavesy[i].kod);
    }
}
//---------------------------------------------------------------------------


void __fastcall TfrmOsd::aeUdalostiMessage(tagMSG &Msg, bool &Handled)
{
    // udalost - klavesova skratka
    if (Msg.message == WM_HOTKEY) {
        int atom = (int) Msg.wParam; // unikatny identifikator
        unsigned char kod = (unsigned int) HIWORD(Msg.lParam); // Virtual-key
        String hlavne; // hlavny informacny text - to, co sa zmenilo
        String ine1, ine2; // stav ostatnych klavesov

        // ak je klaves skratkou, neprevedie sa bezna akcia (napr. stisenie
        // hlasitosti) automaticky, je nutne simulovat ju
           Stlac(kod, atom);

        // zistenie informacii, tvorba informacneho textu
        switch (kod) {
            case VK_CAPITAL:
                hlavne = InfoLock(VK_CAPITAL);
                ine1 = InfoLock(VK_NUMLOCK);
                ine2 = InfoLock(VK_SCROLL);
                break;
            case VK_NUMLOCK:
                hlavne = InfoLock(VK_NUMLOCK);
                ine1 = InfoLock(VK_CAPITAL);
                ine2 = InfoLock(VK_SCROLL);
                break;
            case VK_SCROLL:
                hlavne = InfoLock(VK_SCROLL);
                ine1 = InfoLock(VK_CAPITAL);
                ine2 = InfoLock(VK_NUMLOCK);
                break;
            case VK_VOLUME_DOWN:
            case VK_VOLUME_UP:
                hlavne = HLASITOST + IntToStr(ZistiHlas(HLAS_HLASITOST)) + "%";
                ine1 = (ZistiHlas(HLAS_STLMENIE) ? STLM_ZAP : STLM_VYP);
                break;
            case VK_VOLUME_MUTE:
                hlavne = (ZistiHlas(HLAS_STLMENIE) ? STLM_ZAP : STLM_VYP);
                ine1 = HLASITOST + IntToStr(ZistiHlas(HLAS_HLASITOST)) + "%";
                break;
        }

        // vypisanie informacie, zobrazenie formulara
        lblHlavne->Caption = hlavne;
        lblIne1->Caption = ine1;
        lblIne2->Caption = ine2;
        if (!Application->Active)
            predch_okno = GetForegroundWindow();
        if (!frmOsd->Visible)
            frmOsd->Show();
        tmrOsd->Enabled = false; // restart casovaca
        tmrOsd->Enabled = true; 
    }
}
//---------------------------------------------------------------------------

void __fastcall TfrmOsd::tmrOsdTimer(TObject *Sender)
{
    Hide();
    tmrOsd->Enabled = false;
}
//---------------------------------------------------------------------------

void __fastcall TfrmOsd::aeUdalostiActivate(TObject *Sender)
{
    SetForegroundWindow(predch_okno);
}
//---------------------------------------------------------------------------

void __fastcall TfrmOsd::lblHlavneMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    Hide();
}
//---------------------------------------------------------------------------

void __fastcall TfrmOsd::FormMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
    Hide();
}
//---------------------------------------------------------------------------

void __fastcall TfrmOsd::pmiKoniecClick(TObject *Sender)
{
    Application->Terminate();  
}
//---------------------------------------------------------------------------


